using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;


namespace neorbe
{
	public class DnaSimulator
	{
		public const byte I = (byte)'I';
		public const byte C = (byte)'C';
		public const byte F = (byte)'F';
		public const byte P = (byte)'P';
		readonly StreamWriter writer;
		Dna dna;

		public DnaSimulator(Dna dna, StreamWriter writer)
		{
			this.dna = dna;
			this.writer = writer;
		}

		public void Simulate()
		{
			Stopwatch mainWatch = new Stopwatch();
			Stopwatch sw = new Stopwatch();
			sw.Start();
			mainWatch.Start();
			int count = 0;
			try
			{
				while (true)
				{
					List<PatternItem> pattern = Pattern();
					List<TemplateItem> templ = Template();
					MatchReplace(pattern, templ);
					++count;
					if ((count % 10000) == 0)
					{
						sw.Stop();
						Console.Error.WriteLine("Iterations: " + count);
						Console.Error.WriteLine(sw.ElapsedMilliseconds);
						sw.Reset();
						sw.Start();
					}
				}
			}
			catch (DnaException)
			{
			}
			mainWatch.Stop();
			Console.Error.WriteLine("Total iterations: {0}", count);
			Console.Error.WriteLine("Total time: {0}", mainWatch.Elapsed);
			Console.Error.WriteLine(
				"Average speed: {0}",
				count / (double)mainWatch.ElapsedMilliseconds * 1000);
		}

		List<PatternItem> Pattern()
		{
			int level = 0;
			List<PatternItem> pat = new List<PatternItem>(16);
			while (true)
			{
				if (!dna.Has())
					throw new DnaException();
				switch (dna.GetFirst())
				{
					case C:
						dna = dna.Move();
						pat.Add(new PatternItem.Item(I));
						break;
					case F:
						dna = dna.Move();
						pat.Add(new PatternItem.Item(C));
						break;
					case P:
						dna = dna.Move();
						pat.Add(new PatternItem.Item(F));
						break;
					default:
						if (!dna.Has())
							throw new DnaException();
						Dna dna1 = dna.Move();
						switch (dna1.GetFirst())
						{
							case C:
								dna = dna1.Move();
								pat.Add(new PatternItem.Item(P));
								break;
							case P:
								dna = dna1.Move();
								pat.Add(new PatternItem.Skip(Natural()));
								break;
							case F:
								dna = dna1.Move(2);
								pat.Add(new PatternItem.Sequence(Constants()));
								break;
							default:
								if (!dna1.Has())
									throw new DnaException();
								Dna dna2 = dna1.Move();
								switch (dna2.GetFirst())
								{
									case P:
										dna = dna2.Move();
										++level;
										pat.Add(new PatternItem.Open());
										break;
									case C:
									case F:
										dna = dna2.Move();
										if (level == 0)
											return pat;
										--level;
										pat.Add(new PatternItem.Close());
										break;
									default:
										dna = dna2.Move();
										foreach (byte b in dna.Take(7))
											writer.Write((char)b);
										dna = dna.Move(7);
										break;
								}
								break;
						}
						break;
				}
			}
		}

		int Natural()
		{
			int n = 0;
			int b = 1;
			while (true)
			{
				if (!dna.Has())
					throw new DnaException();
				switch (dna.GetFirst())
				{
					case P:
						dna = dna.Move(1);
						return n;
					case C:
						if (b < 0x3FFFFFFF)
							n += b;
						break;
					default:
						break;
				}
				dna = dna.Move();
				if (b < 0x3FFFFFFF)
					b *= 2;
			}
		}

		byte[] Constants()
		{
			List<byte> result = new List<byte>(16);
			Dna dna = this.dna;
			while (true)
			{
				if (!dna.Has())
					goto exit;
				switch (dna.GetFirst())
				{
					case C:
						dna = dna.Move();
						result.Add(I);
						break;
					case F:
						dna = dna.Move();
						result.Add(C);
						break;
					case P:
						dna = dna.Move();
						result.Add(F);
						break;
					//I
					default:
						Dna d = dna.Move();
						if (!d.Has() || d.GetFirst() != C)
							goto exit;
						result.Add(P);
						dna = d.Move();
						break;
				}
			}
		exit:
			this.dna = dna;
			return result.ToArray();
		}

		void FlushTemplateBuffer(List<TemplateItem> templ, List<byte> buffer)
		{
			if (buffer.Count == 0)
				return;
			templ.Add(new TemplateItem.Items(buffer.ToArray()));
			buffer.Clear();
		}

		List<TemplateItem> Template()
		{
			List<TemplateItem> templ = new List<TemplateItem>(16);
			List<byte> buffer = new List<byte>(16);
			while (true)
			{
				if (!dna.Has())
					throw new DnaException();
				switch (dna.GetFirst())
				{
					case C:
						dna = dna.Move();
						buffer.Add(I);
						break;
					case F:
						dna = dna.Move(1);
						buffer.Add(C);
						break;
					case P:
						dna = dna.Move(1);
						buffer.Add(F);
						break;
					default:
						if (!dna.Has())
							throw new DnaException();
						Dna dna1 = dna.Move();
						switch (dna1.GetFirst())
						{
							case C:
								dna = dna1.Move();
								buffer.Add(P);
								break;
							case F:
							case P:
								{
									FlushTemplateBuffer(templ, buffer);
									dna = dna1.Move();
									int l = Natural();
									int n = Natural();
									templ.Add(new TemplateItem.Protected(n, l));
								}
								break;
							default:
								if (!dna.Has())
									throw new DnaException();
								Dna dna2 = dna1.Move();
								switch (dna2.GetFirst())
								{
									case C:
									case F:
										dna = dna2.Move();
										FlushTemplateBuffer(templ, buffer);
										return templ;
									case P:
										{
											FlushTemplateBuffer(templ, buffer);
											dna = dna2.Move();
											int n = Natural();
											templ.Add(new TemplateItem.Number(n));
										}
										break;
									default:
										dna = dna2.Move();
										foreach (byte b in dna.Take(7))
											writer.Write((char)b);
										dna = dna.Move(7);
										break;
								}
								break;
						}
						break;
				}
			}
		}

		private struct DnaAndSize
		{
			public readonly Dna Dna;
			public readonly int Size;

			public DnaAndSize(Dna dna, int size)
			{
				this.Dna = dna;
				this.Size = size;
			}
		}

		void MatchReplace(List<PatternItem> pat, List<TemplateItem> templ)
		{
			List<Dna> env = new List<Dna>(16);
			Dna i = dna;
			int sz = 0;
			Stack<DnaAndSize> stack = new Stack<DnaAndSize>(16);
			foreach (PatternItem p in pat)
			{
				switch (p.Code)
				{
					case PatternItem.ItemCode:
						{
							PatternItem.Item item = (PatternItem.Item)p;
							if (i.Has() && i.GetFirst() == item.Value)
							{
								++sz;
								i = i.Move();
							}
							else
								return;
						}
						break;
					case PatternItem.SkipCode:
						{
							PatternItem.Skip skip = (PatternItem.Skip)p;
							bool has = i.Has(skip.Size);
							i = i.Move(skip.Size);
							if (!has)
								return;
							sz += skip.Size;
						}
						break;
					case PatternItem.SequenceCode:
						{
							PatternItem.Sequence seq = (PatternItem.Sequence)p;
							int findSize;
							i = Find(i, seq.Value, out findSize);
							if (findSize < 0)
								return;
							sz += findSize;
						}
						break;
					case PatternItem.OpenCode:
						stack.Push(new DnaAndSize(i, sz));
						break;
					default:
						{
							DnaAndSize inf = stack.Pop();
							env.Add(inf.Dna.GetSlice(sz - inf.Size));
						}
						break;
				}
			}
			dna = i;
			Replace(templ, env);
		}

		void FlushReplaceBuffer(List<byte[]> buffer, List<Dna> dnas)
		{
			if (buffer.Count == 0)
				return;
			int size = 0;
			foreach (byte[] block in buffer)
				size += block.Length;
			byte[] result = new byte[size];
			int offset = 0;
			foreach (byte[] block in buffer)
			{
				Array.Copy(block, 0, result, offset, block.Length);
				offset += block.Length;
			}
			dnas.Add(new Dna(result, Dna.Empty));
			buffer.Clear();
		}

		void Replace(List<TemplateItem> templ, List<Dna> env)
		{
			List<Dna> r = new List<Dna>(16);
			List<byte[]> buffer = new List<byte[]>(16);
			foreach (TemplateItem t in templ)
			{
				switch (t.Code)
				{
					case TemplateItem.ItemsCode:
						{
							TemplateItem.Items items = (TemplateItem.Items)t;
							buffer.Add(items.Value);
						}
						break;
					case TemplateItem.NumberCode:
						{
							TemplateItem.Number num = (TemplateItem.Number)t;
							int sz = num.Reference >= 0 && num.Reference < env.Count ?
								env[num.Reference].Size : 0;
							buffer.Add(EncodeNatural(sz));
						}
						break;
					case TemplateItem.ProtectedCode:
						{
							TemplateItem.Protected prot = (TemplateItem.Protected)t;
							if (prot.Reference >= 0 && prot.Reference < env.Count)
							{
								if (prot.Protection == 0 && env[prot.Reference].Size >= 512)
								{
									FlushReplaceBuffer(buffer, r);
									r.Add(env[prot.Reference]);
								}
								else
								{
									buffer.Add(Protect(prot.Protection, env[prot.Reference].Take()));
								}
							}
						}
						break;
				}
			}
			FlushReplaceBuffer(buffer, r);
			dna = dna.AppendToStart(Dna.Concat(r));
		}

		byte[] Protect(int count, IEnumerable<byte> seq)
		{
			List<byte> result = new List<byte>(seq);
			for (int i = 0; i < count; ++i)
			{
				List<byte> tmp = new List<byte>(16);
				foreach (byte b in result)
				{
					switch (b)
					{
						case I:
							tmp.Add(C);
							break;
						case C:
							tmp.Add(F);
							break;
						case F:
							tmp.Add(P);
							break;
						default:
							tmp.Add(I);
							tmp.Add(C);
							break;
					}
				}
				result = tmp;
			}
			return result.ToArray();
		}

		byte[] EncodeNatural(int n)
		{
			List<byte> result = new List<byte>(256);
			while (n > 0)
			{
				if (n % 2 == 0)
					result.Add(I);
				else
					result.Add(C);
				n /= 2;
			}
			result.Add(P);
			return result.ToArray();
		}

		static Dna Find(Dna dna, byte[] seq, out int sz)
		{
			List<int> automata = new List<int>(16);
			int i = 0;
			while (dna != Dna.Empty)
			{
				automata.Add(0);
				int j = 0;
				byte b = dna.GetFirst();
				while (j < automata.Count)
				{
					int state = automata[j];
					if (state >= seq.Length)
					{
						sz = i;
						return dna;
					}
					else
					{
						if (b == seq[state])
						{
							++automata[j];
							++j;
						}
						else
							automata.RemoveAt(j);
					}
				}
				dna = dna.Move();
				++i;
			}
			foreach (int state in automata)
			{
				if (state >= seq.Length)
				{
					sz = i;
					return dna;
				}
			}
			sz = -1;
			return dna;
		}
	}
}
