using System;
using System.Collections.Generic;
using System.IO;
using System.Text;


namespace neorbe
{
	public class Dna
	{
		static int avgBlockSize = 256;
		public static readonly Dna Empty = new Dna(null, 0, 0, null);

		private readonly byte[] data;
		private readonly int first;
		private readonly int last;
		private readonly Dna next;
		private int size = -1;
		private int blocksCount = -1;

		public Dna(byte[] data, int first, int last, Dna next)
		{
			this.data = data;
			this.first = first;
			this.last = last;
			this.next = next;
		}

		public Dna(byte[] data, Dna next)
			: this(data, 0, data.Length, next)
		{
		}

		public byte GetFirst()
		{
			unchecked
			{
				return data[first];
			}
		}

		public bool Has()
		{
			return this != Empty;
		}

		public int Size
		{
			get
			{
				if (size == -1)
				{
					size = last - first;
					if (next != null)
						size += next.Size;
				}
				return size;
			}
		}

		public int BlocksCount
		{
			get
			{
				if (blocksCount == -1)
				{
					blocksCount = 1;
					if (next != null)
						blocksCount += next.BlocksCount;
				}
				return blocksCount;
			}
		}

		public bool Has(int count)
		{
			if (this == Empty)
				return false;
			Dna current = this;
			while (current.first + count > current.last)
			{
				count -= current.last - current.first;
				current = current.next;
				if (current == Empty)
					return false;
			}
			return true;
		}

		public Dna Move()
		{
			if (this == Empty)
				return Empty;
			if (first + 1 == last)
				return next;
			else
				return new Dna(data, first + 1, last, next);
		}

		public Dna Move(int offset)
		{
			if (this == Empty)
				return Empty;
			Dna current = this;
			while (current.first + offset >= current.last)
			{
				offset -= current.last - current.first;
				current = current.next;
				if (current == Empty)
					return current;
			}
			return new Dna(current.data, current.first + offset, current.last, current.next);
		}

		public Dna GetSlice(int count)
		{
			if (this == Empty)
				return this;
			Stack<Dna> stack = new Stack<Dna>(16);
			Dna current = this;
			while (current.first + count >= current.last)
			{
				count -= current.last - current.first;
				stack.Push(current);
				current = current.next;
				if (current == Empty)
					return this;
			}
			if (count > 0)
				current = new Dna(current.data, current.first, current.first + count, Empty);
			else
				current = Dna.Empty;
			while (stack.Count > 0)
			{
				Dna dna = stack.Pop();
				current = new Dna(dna.data, dna.first, dna.last, current);
			}
			return current;
		}

		public Dna AppendToStart(Dna prefix)
		{
			if (this == Empty)
				return prefix;
			Stack<Dna> stack = new Stack<Dna>(16);
			Dna current = prefix;
			while (current != Empty)
			{
				stack.Push(current);
				current = current.next;
			}
			current = this;
			while (stack.Count > 0)
			{
				Dna dna = stack.Pop();
				current = new Dna(dna.data, dna.first, dna.last, current);
			}
			if (current.BlocksCount > avgBlockSize)
			{
				avgBlockSize += 54;
				if (avgBlockSize > 280)
				{
					avgBlockSize = 50;
					current = current.Defragment(130072);
				}
				else
				{
					current = current.Defragment(6000);
				}
			}
			else
			{
				if (avgBlockSize > 0)
					--avgBlockSize;
			}
			
			return current;
		}

		public Dna Defragment(int avg)
		{
			unchecked
			{
				List<ArraySegment<byte>> buffer = new List<ArraySegment<byte>>();
				Dna current = this;
				Stack<Dna> stack = new Stack<Dna>(16);
				while (current != Empty)
				{
					if (current.last - current.first <= avg)
					{
						buffer.Add(new ArraySegment<byte>(
							current.data, current.first, current.last - current.first));
					}
					else
					{
						if (buffer.Count > 0)
						{
							stack.Push(new Dna(Utils.Concat(buffer), Empty));
							buffer.Clear();
						}
						stack.Push(current);
					}
					current = current.next;
				}
				if (buffer.Count > 0)
					current = new Dna(Utils.Concat(buffer), Empty);
				else
					current = Empty;
				while (stack.Count > 0)
				{
					Dna dna = stack.Pop();
					current = new Dna(dna.data, dna.first, dna.last, current);
				}
				return current;
			}
		}

		public IEnumerable<byte> Take(int count)
		{
			unchecked
			{
				Dna current = this;
				while (current != Empty)
				{
					for (int i = current.first; i < current.last; ++i)
					{
						if (count <= 0)
							yield break;
						--count;
						yield return current.data[i];
					}
					current = current.next;
				}
			}
		}

		public IEnumerable<byte> Take()
		{
			unchecked
			{
				Dna current = this;
				while (current != Empty)
				{
					for (int i = current.first; i < current.last; ++i)
						yield return current.data[i];
					current = current.next;
				}
			}
		}

		public string TakeAsString(int count)
		{
			unchecked
			{
				StringBuilder sb = new StringBuilder();
				foreach (byte b in Take(count))
					sb.Append((char)b);
				return sb.ToString();
			}
		}

		public static Dna Concat(IList<Dna> dnas)
		{
			Dna result = Dna.Empty;
			for (int i = dnas.Count - 1; i >= 0; --i)
			{
				Dna current = dnas[i];
				if (current == Empty)
					continue;
				Stack<Dna> stack = new Stack<Dna>();
				while (current != Empty)
				{
					stack.Push(current);
					current = current.next;
				}
				current = result;
				while (stack.Count > 0)
				{
					Dna dna = stack.Pop();
					current = new Dna(dna.data, dna.first, dna.last, current);
				}
				result = current;
			}
			return result;
		}
	}
}
